(* ********************************************************* *)
(*                                                           *)
(* FUNNEL  Version  June 2, 2023          Free Pascal      *)
(*         Brian Fristensky                                  *)
(*         Dept. of Plant Science                            *)
(*         University of Manitoba                            *)
(*         Winnipeg, MB R3T 2N2  CANADA                      *)
(*                                                           *)
(* Copyright (c) 1987, 1990 by Brian Fristensky              *)
(* !!! in comment indicates feature which may need change    *)
(* ********************************************************* *)
program FUNNEL(input,output (*,INFILE,OUTFILE*));
(*!!! Some Pascals require file parameters in program heading *)
 

uses 
      Sysutils, Crt;


const VERSION = 'FUNNEL         Version  June 3, 2023';
      MAXWORD = 25;
      MAXLINE = 10000;
(* BEGIN MODULE INTLIMITS *)
(*!!!*)  MAXINT =  2147483647;
(*!!! *) MININT = -2147483647;
(* END MODULE INTLIMITS *)
 
type
(* BEGIN MODULE TYPE.WORD *)
       (*   <word>::= <non-blank char>[<non-blank char>] *)
       WORD    = record
                 LEN:integer;
                 STR:array[1..MAXWORD] of char
                 end;
(* END MODULE TYPE.WORD *)

(* BEGIN MODULE TYPE.LINE *)
       CHARARRAY = array[1..MAXLINE] of char;
       LINE = record
                INPSTR:string;
                STR:CHARARRAY;
                LEN:integer
                end;
(* END MODULE TYPE.LINE *)
 
var INFILE,OUTFILE:text;
    IFN,OFN:LINE;
    CH: char;
    I,LINESIZE: longint;
    COMMENT     : boolean;

(* BEGIN MODULE INPLINE *) 
    (*  Read a line from the console.    *)
    procedure INPLINE(var W:LINE);
      var I : integer;
      begin
        with W do begin
            readln(input,INPSTR);
            I := 1;
            LEN := Length(INPSTR);
            while I <= LEN do begin
                STR[I] := INPSTR[I];
                I := I + 1
            end
        end
      end; (* INPLINE *)
(* END MODULE INPLINE *)
 
(* BEGIN MODULE GETFILE *)

(************************************************************)
(* Open  files, checking for I/O errors.                    *)
(* Assumes user has already been prompted for filename.     *)
(* Syntax: GETFILE(F,'I',<filename>);                       *)
(*              opens file F for input                      *)
(*         GETFILE(F,'O',<filename>);                       *)
(*              opens file F for output                     *)
(* !!! NOTE: this non-Standard procedure must be rewritten  *)
(* for each different version of Pascal.                    *)
(************************************************************)
procedure GETFILE(var F:text; FTYPE:char; var FILENAME:LINE);

  const MAXATTEMPTS = 5;

  var CH:char;
      OKAY:boolean;
      ATTEMPTS : integer;

  begin (* GETFILE - - - - - - - - - - - - - - - - - - - - - *)
      (* Read in filename.*)
      INPLINE(FILENAME);
      ATTEMPTS := 1;
      case FTYPE of
        (* Input file must exist before being opened.*)
        'I':begin (* input file *)
              while not(FileExists(FILENAME.INPSTR)) do begin
                writeln('***FILE NOT FOUND');
                writeln('Enter another filename:');
                (* Prevent an infinite loop. *)
                if ATTEMPTS > MAXATTEMPTS then begin
                    writeln('>>> Too many attempts. Terminating program.');
                    Halt
                    end
                else begin
                    INPLINE(FILENAME);
                    ATTEMPTS := ATTEMPTS + 1
                    end
                end; (* while *)
            Assign(F,FILENAME.INPSTR);
(*!!!*)     reset(F)
            end; (* I *)

        (* If output file already exists, ask user if he/she/they really want
           to overwrite it. *)
        'O':begin
              repeat
                OKAY:=true;
                if FileExists(FILENAME.INPSTR) then begin
                  repeat  
                    writeln('*** WARNING! File already exists. Overwrite?[Y|N]');
                    readln(CH);
                    if not ( CH in ['Y','y']) then begin
                        (* Prevent an infinite loop. *)
                        if ATTEMPTS > MAXATTEMPTS then begin
                             writeln('>>> Too many attempts. Terminating program.');
                            Halt
                            end
                        else begin
                            writeln('Enter another filename:');
                            INPLINE(FILENAME);
                            ATTEMPTS := ATTEMPTS + 1;
                            OKAY := false
                            end
                        end (* not CH in ['Y','y'] *)
                  until CH in ['Y','y','N','n']
                  end (* EXISTS(FILENAME) *)
                else FileCreate(FILENAME.INPSTR)
              until OKAY;
            Assign(F,FILENAME.INPSTR);
            rewrite(F)  (* output file *)
            end (* O *)
        end (*case*)
  end;  (* GETFILE *)
(* END MODULE GETFILE *)
 
(* BEGIN MODULE INPWORD *)
  (* Read a WORD from the terminal. Does NOT read the rest of the line. *)
   procedure INPWORD(var W:WORD);
      var I : integer;
         CH : char;
         BACKSPACE: char;
      begin
        with W do begin
          BACKSPACE:= chr(8);
          LEN:= 0;
          if not(eoln(input)) then read(input,CH);
          while (CH = ' ') do read(input,CH);
          LEN := 1;
          STR[LEN] := CH;
          while (CH <> ' ') and not(eoln(input)) do begin
            read(input,CH);
            if (CH= BACKSPACE) and (LEN > 0) then begin
              write(BACKSPACE);LEN:= LEN-1 end
            else if LEN < MAXWORD then begin
              LEN := LEN + 1;
              STR[LEN]:= CH
            end 
          end;
        if eoln(input) then readln;
        for I := LEN+1 to MAXWORD do STR[I] := ' '
        end
      end; (* INPWORD *)
(* END MODULE INPWORD *)
 
(* BEGIN MODULE GETINTEGER *)
  (* Prompts user for an integer, checks whether all characters are digits,*)
  (* and whether number is within desired range; harasses user until valid *)
  (* integer is received. *)
  procedure GETINTEGER(var NUM:longint;LBOUND,HBOUND:longint);
 
   const
     MAXATTEMPTS = 3;
   var I,VAL,ORDZERO,ATTEMPTS  : integer;
     LEGAL,INRANGE,NEGATIVE      : boolean;
     CH : char;
     NUMWORD : WORD;
 
  begin
    INRANGE:= false;
    LEGAL := false;
    NEGATIVE := false;
    ORDZERO:= ord('0');
    ATTEMPTS := 0;
    repeat
      ATTEMPTS := 0;
      repeat
        INPWORD(NUMWORD);
        with NUMWORD do begin
 
          (* Evaluate sign, if any *)
          I:=1;CH:=STR[I];
          if CH = '-' then begin
             NEGATIVE := true;
             I:=I+1; CH:= STR[I]
             end
          else begin NEGATIVE:= false; if CH='+' then begin
                         I:=I+1; CH:= STR[I] end end;
 
          (* Evaluate unsigned integer *)
          NUM:= 0;
          while CH in ['0'..'9'] do
            begin
              VAL:= ord(CH)-ORDZERO;
              NUM:= (NUM * 10) + VAL;
              I:= I+1; CH:= STR[I]
            end;
          if I > LEN then LEGAL:= true
          else  begin
              LEGAL:= false;
              writeln;
              writeln('Illegal character encountered.');
              writeln('Enter new value:  ');
             end
         end; (* with *)
         ATTEMPTS := ATTEMPTS + 1
       until LEGAL or (ATTEMPTS > MAXATTEMPTS);
       if ATTEMPTS > MAXATTEMPTS then begin
           writeln('Too many attempts. Program terminating');
           Halt
           end;
       (* writeln(NUM); *)
       (* If the number entered was negative, multiply *)
       (* NUM by -1. Check range of number.            *)
       if NEGATIVE then NUM:= - NUM;
       if (NUM >= LBOUND) and (NUM <= HBOUND)
         then INRANGE := true
       else begin
         INRANGE := false;
         writeln;
         writeln('Number is out of range.');
         writeln('Please enter new value:')
         end;
       ATTEMPTS := ATTEMPTS + 1
     until INRANGE or (ATTEMPTS > MAXATTEMPTS);
    if (ATTEMPTS > MAXATTEMPTS) then begin
         writeln('Too many attempts. Program terminating');
         Halt
    end
   end;   (* GETINTEGER *)
(* END MODULE GETINTEGER *)
 
begin
(* BEGIN MODULE STARTUP *)
(* Peform operations which must be done at beginning of main
   procedure. *)
(*!!!   TERMIN(input);    Open input for interactive use *)
(*!!!   TERMOUT(output);   "   output "      "        "  *)
     writeln(VERSION:50);
     writeln;
(* END MODULE STARTUP *)
(*Ask user for INFILE and OUTFILE filenames*)
  writeln('INPUT FILENAME?');
  GETFILE(INFILE,'I',IFN);
  writeln('OUTPUT FILENAME?');
  GETFILE(OUTFILE,'O',OFN);
  writeln('Enter LINESIZE:');
  GETINTEGER(LINESIZE,1,MAXLINE);readln;
  (*READ INPUT FILE CHARACTER BY CHARACTER*)
  (*AND WRITE EACH CHARACTER INTO OUTPUT FILE*)
  I:= 0;COMMENT:=false;
  while not eof(INFILE) do begin
      if COMMENT then I:= 0;
      COMMENT:= false;
      while not eoln(INFILE) do begin
          read(INFILE,CH);
          if CH in [';','>'] then begin
                           COMMENT := true;
                           if I>0 then writeln(OUTFILE) 
                           end;
          case COMMENT of
            true: write(OUTFILE,CH);
            false:if CH in ['A'..'Z','a'..'z','*']
             then begin
                    write(OUTFILE,CH);
                    I:= I + 1;
                    if I = LINESIZE then  begin
                      I:=0; writeln(OUTFILE) end
                    end
              end
          end;
      readln(INFILE);
      if COMMENT then writeln(OUTFILE) 
      end;
 (*!!!*)CLOSE(OUTFILE) 
  end. (* FUNNEL *)
 
